#     Copyright 2018 Netflix, Inc.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
"""
.. module: policy_parrot.action
    :platform: Unix

.. version:: $$VERSION$$
.. moduleauthor::  Patrick Kelley <patrickbarrettkelley@gmail.com> @patrickbkelley

"""

import fnmatch

from policy_parrot import all_permissions
from policy_parrot.common import ensure_array


def build_service_actions_from_service_data(service_data):
    permissions = set()
    for service_name in service_data:
        prefix = service_data[service_name]["prefix"]
        service_actions = service_data[service_name]["actions"]
        for action in service_actions:
            permissions.add("{}:{}".format(prefix, action.lower()))
    return permissions


def _expand(action):
    """
    :param action: 'autoscaling:*'
    :return: A list of all autoscaling permissions matching the wildcard
    """
    expanded = fnmatch.filter(all_permissions, action.lower())
    # if we get a wildcard for a tech we've never heard of, just return the wildcard
    if not expanded:
        return [action]
    return expanded


def _expand_wildcard_action(actions):
    """Expand wildcards in a list of actions (or a single action string), returning a list of all matching actions.

    :param actions: ['autoscaling:*']
    :return: A list of all permissions matching the action globs
    """
    if isinstance(actions, str):
        # Bail early if we have a string with no wildcard
        if "*" not in actions:
            return [actions.lower()]
        actions = [actions]

    # Map _expand function to action list, resulting in a list of lists of expanded actions.
    temp = map(_expand, actions)

    # This flattens the list of lists. It's hard to read, but it's a hot path and the optimization
    # speeds it up by 90% or more.
    expanded = [item.lower() for sublist in temp for item in sublist]

    return expanded


def get_actions_from_statement(statement):
    allowed_actions = set()
    actions = ensure_array(statement.get("Action", []))

    for action in actions:
        allowed_actions = allowed_actions.union(set(_expand_wildcard_action(action)))

    inverted_actions = set()
    not_actions = ensure_array(statement.get("NotAction", []))

    for action in not_actions:
        inverted_actions = inverted_actions.union(set(_expand_wildcard_action(action)))

    if inverted_actions:
        actions = _invert_actions(inverted_actions)
        allowed_actions = allowed_actions.union(actions)

    return allowed_actions


def _invert_actions(actions):
    from policy_parrot import all_permissions

    return all_permissions.difference(actions)
